/*
 * CImagenexSidescanSonar.cpp
 *
 *  Created on: Mar 2, 2011
 *      Author: Joel Vidal
 */

#include "CImagenexSidescanSonar.h"


CImagenexSidescanSonar::CImagenexSidescanSonar( const std::string& componentName )
:
	CComponent( componentName ),
	thread_() ,
	device_( new CTcpSocket ) ,
	imagenexSidescanSonarDataManip_() ,
	configDataManip_( getConfigTagName() )
{

}



void
CImagenexSidescanSonar::initialize()
	throw ( std::runtime_error )
{
	setConfiguration( station().getConfiguration( getName() ) ) ;

	try {
		boost::dynamic_pointer_cast< CTcpSocket >( device_ )->
		connect( configDataManip_.data().deviceIP, configDataManip_.data().devicePort ) ;

		if ( configDataManip_.data().imageCreate )
			imageInitializeView() ;

		logger().information( imagenexSidescanSonarDataManip_.getLogHeader() ) ;

		thread_ = threadFactory().createThread( IThread::MODE_NONPERIODIC ) ;
		thread_->start( boost::bind( &CImagenexSidescanSonar::run, this) ) ;
	}
	catch ( std::exception& e ) {
		logger().error( std::string( "Initialize - " ) + e.what() ) ;
	}


}



std::string
CImagenexSidescanSonar::getConfiguration() const
	throw ( std::runtime_error )
{
	return configDataManip_.getXmlString() ;
}



void
CImagenexSidescanSonar::setConfiguration( const std::string& configuration )
	throw ( std::runtime_error )
{
	configDataManip_.loadXmlString( configuration ) ;
}



void
CImagenexSidescanSonar::run()
{
	int packetNumber = 0 ;

	while ( true ) {
		if ( isEnabled() ) {
			try {
				//Enviem petició
				sendSDC( packetNumber ) ;

				//Esparem dades
				recieveIRD( packetNumber ) ;

				if ( packetNumber == 0 ) packetNumber = 2 ;
				else packetNumber = 0 ;
			}
			catch ( IIODevice::ReadTimeout& e ) {
				logger().warning( "I/O device reading timeout" ) ;
				packetNumber = 0 ;
			}
			catch ( std::exception& e ) { // Could be std::runtime_error, but will use defensive programming instead
				logger().error( e.what() );
				packetNumber = 0 ;
			}
		}
	}
}



void
CImagenexSidescanSonar::sendSDC( unsigned char packetNumber )
{
	if ( packetNumber != 0 && packetNumber != 2  ) {
		logger().warning( "Packet Number error" ) ;
		packetNumber = 0 ;
	}

	//Get parameters
	int range = configDataManip_.data().range ;
	int freq = configDataManip_.data().frequency ;
	int dataGain = configDataManip_.data().dataGain ;
	int balanceGain = configDataManip_.data().balanceGain ;

	//Check data_gain parameter
	if ( dataGain < 0 ) {
		dataGain = 0 ;
		logger().warning( std::string( "data_gain - out of rang" ) ) ;
	}
	else if ( dataGain > 40 ) {
		dataGain = 40 ;
		logger().warning( std::string( "data_gain - out of rang" ) ) ;
	}

	//Check balance_gain parameter
	if ( balanceGain < 0 ) {
		balanceGain = 0 ;
		logger().warning( std::string( "balance_gain - out of rang" ) ) ;
	}
	else if ( balanceGain > 60 ) {
		balanceGain = 60 ;
		logger().warning( std::string( "balance_gain - out of rang" ) ) ;
	}

	//Check frequency parameter
	if ( freq < 0 ) {
		freq = 0 ;
		logger().warning( std::string( "frequency - out of rang" ) ) ;
	}
	else if ( freq > 2 ) {
		freq = 2 ;
		logger().warning( std::string( "frequency - out of rang" ) ) ;
	}

	//Check range parameter
	if ( range < 10 ) {
		range = 10 ;
		logger().warning( std::string( "range - out of rang" ) ) ;
	}
	else {
		if ( freq == 2 ) {
			if ( range > 50 ) {
				range = 50 ;
				logger().warning( std::string( "range - out of rang" ) ) ;
			}
		}
		else {
			if ( range > 200 ) {
				range = 200 ;
				logger().warning( std::string( "range - out of rang" ) ) ;
			}
		}
	}

	//Switch Data Command
	IIODevice::DataBuffer sdc( 27 );

	sdc[0]	= 	0xFE ;			//Byte0, 0xFE
	sdc[1]	= 	0x44 ;			//Byte1, 0x44
	sdc[2]	= 	0 ;				//Byte2, Reserved
	sdc[3]	= 	range ;			//Byte3, Range
	sdc[4]	= 	0 ;				//Byte4, Reserved
	sdc[5]	= 	0 ;				//Byte5, Reserved
	sdc[6]	= 	0 ;				//Byte6, Mater/Slave
	sdc[7]	= 	freq ;			//Byte7, Frequency
	sdc[8]	= 	dataGain ;		//Byte8, Data Gain
	sdc[9]	= 	0 ;				//Byte9,  Reserved
	sdc[10]	= 	balanceGain ;	//Byte10, Balance Gain
	sdc[11]	= 	0 ;				//Byte11, Reserved
	sdc[12]	= 	0 ;				//Byte12, Reserved
	sdc[13]	= 	0 ;				//Byte13, Reserved
	sdc[14]	= 	0 ;				//Byte14, Reserved
	sdc[15]	= 	0 ;				//Byte15, Reserved
	sdc[16]	= 	0 ;				//Byte16, Reserved
	sdc[17]	= 	0 ;				//Byte17, Reserved
	sdc[18]	= 	packetNumber ;	//Byte18, Packet Number
	sdc[19]	= 	0 ;				//Byte19, Reserved
	sdc[20]	= 	0 ;				//Byte20, Reserved
	sdc[21]	= 	0 ;				//Byte21, Reserved
	sdc[22] = 	0 ;				//Byte22, Reserved
	sdc[23]	= 	0 ;				//Byte23, Reserved
	sdc[24]	= 	0 ;				//Byte24, Reserved
	sdc[25]	= 	0 ;				//Byte25, Reserved
	sdc[26]	= 	0xFD ;			//Byte26, 0xFD

	//Send sdc
	device_->write( sdc ) ;
}



void
CImagenexSidescanSonar::recieveIRD( unsigned char packetNumber )
{

	if ( packetNumber != 0 && packetNumber != 2  ) {
		logger().warning( "Packet Number error" ) ;
		packetNumber = 0 ;
	}

	IIODevice::DataBuffer ird ;

	//Read Interface return data
	device_->read( ird, 1013, configDataManip_.data().deviceTimeout ) ;


	if (ird[ 0 ] != 'I' || ird[ 1 ] != 'V' || ird[ 2 ] != 'X' || ird[ 1012 ] != 0xFC )
		logger().warning( std::string( "Not a Imagenex Sidescan Sonar packet" ) ) ;

	//Check serialStatus
	//( ird[ 3 ] & 0x80 ) >> 7	Character Overrun
	//( ird[ 3 ] & 0x40 ) >> 6	Switches Accepted
	//( ird[ 3 ] & 0x20 ) >> 5	Packet Num Error
	//( ird[ 3 ] & 0x10 ) >> 4	Gain Switch Error
	//( ird[ 3 ] & 0x08 ) >> 3	Freq Switch Error
	//( ird[ 3 ] & 0x04 ) >> 2	Range Switch Error
	//( ird[ 3 ] & 0x02 ) >> 1  Channel Bit 1
	//( ird[ 3 ] & 0x01 ) 		Channel Bit 2

	IIODevice::DataBuffer head( 12 );
	std::copy( ird.begin(), ird.begin() + 12, head.begin() ) ;

	IIODevice::DataBuffer channel( 1000 );
	std::copy( ird.begin() + 12, ird.end() - 1, channel.begin() ) ;


	//Update Logger & Publish
	if (packetNumber == 0 ) {

		imagenexSidescanSonarDataManip_.data().portChannel.resize( 1000 ) ;

		for( unsigned int i = 0 ; i < 1000 ; i++ )
			imagenexSidescanSonarDataManip_.data().portChannel[ i ] =  channel[ i ] ;
	}
	else {

		imagenexSidescanSonarDataManip_.data().starboardChannel.resize( 1000 ) ;

		for( unsigned int i = 0 ; i < 1000 ; i++ )
			imagenexSidescanSonarDataManip_.data().starboardChannel[ i ] =   channel[ i ] ;

		imagenexSidescanSonarDataManip_.data().serialStatus = head[ 3 ] ;
		imagenexSidescanSonarDataManip_.data().range = head[ 4 ] ;
		imagenexSidescanSonarDataManip_.data().frequency = head[ 5 ] ;
		imagenexSidescanSonarDataManip_.data().gain = configDataManip_.getData().dataGain ;

		logger().information( imagenexSidescanSonarDataManip_.getLogEntry() ) ;
		publish(  imagenexSidescanSonarDataManip_.getName(),  imagenexSidescanSonarDataManip_.getXmlString() ) ;
	}


	//Create Image


	 if ( image_.create ) {

		int varTime ;
		Poco::DateTime now ;

		int msTime = now.second() * 1000 + now.millisecond() ;

		//x%m is always in the range [-m+1, m-1], (x%m+m)%m is always in the range [0, m-1]
		varTime = ( ( msTime - image_.iniPingTime ) % 60000 + 60000 ) % 60000 ;

		if ( varTime >= image_.pingTime ) {
			if ( packetNumber == 0 ) {

				IIODevice::DataBuffer  channelResize ;
				imageScaleChannel (  channel, channelResize ) ;
				imageAddChannel( channelResize, packetNumber ) ;
				image_.firstPingEnd = true ;

			}
			else {
				if ( image_.firstPingEnd ) {

					IIODevice::DataBuffer channelResize ;
					imageScaleChannel ( channel, channelResize ) ;
					imageAddChannel( channelResize, packetNumber ) ;
					image_.firstPingEnd = false ;
					image_.iniPingTime = msTime ;

				}
			}
		}


		if ( packetNumber == 2 ) {

			//x%m is always in the range [-m+1, m-1], (x%m+m)%m is always in the range [0,  m-1]
			varTime = ( ( msTime - image_.iniRefreshTime ) % 60000 + 60000 ) % 60000 ;

			if ( varTime >= image_.refreshTime ) {
				imageDraw() ;
				image_.iniRefreshTime = msTime ;
			}
		}

	}

}



void
CImagenexSidescanSonar::imageInitializeView() {

	//Initialize variables
	image_.create = true ;
	image_.width = configDataManip_.data().imageWidthPx ;
	image_.height = configDataManip_.data().imageHeightPx ;
	//image_.refreshTime = 1000 /  configDataManip_.data().imageFrequency ;

	image_.refreshTime = configDataManip_.data().imageRefresh ;
	image_.iniRefreshTime = 0 ;

	image_.pingTime = configDataManip_.data().imagePing ;
	image_.iniPingTime = 0 ;
	image_.firstPingEnd = false ;

	//check width % 2 = 0
	if ( image_.width % 2 != 0  ) {
		logger().warning( "image_width_px should be even number" ) ;
		image_.width++ ;
	}

	//Create circular buffer
	image_.circularBuffer  =  static_cast< char * > ( malloc ( image_.width * image_.height ) ) ;

	//Initialize circular buffer variables
	image_.bufferSize =  image_.width * image_.height ;
	image_.bufferPos = 0 ;

	//Paint Back
	for ( int i = 0 ; i < image_.bufferSize ; i++ ) image_.circularBuffer[ i ] = 255 ;

	//Create Image
	image_.imageView.create( image_.height, image_.width, CV_8UC1 ) ;

	//Create Windows
	cv::namedWindow( "Imaginex Sidescan Sonar", CV_WINDOW_AUTOSIZE ) ;
	cv::waitKey( 100 ) ;

}



void
CImagenexSidescanSonar::imageStopView() {

	//Change variables
	image_.create = false ;

	//Delete circular buffer
	free ( image_.circularBuffer );

	//Delete Image
	image_.imageView.release() ;


	//Delete Windows
	cvDestroyAllWindows() ;

}



void
CImagenexSidescanSonar::imageAddChannel( const IIODevice::DataBuffer & channel,
										 unsigned char packetNumber ) {

	//Add line
	if ( packetNumber != 0 && packetNumber != 2  ) {
		logger().warning( "Packet Number error" ) ;
		packetNumber = 0 ;
	}

	if (packetNumber == 0) {
		std::copy( channel.rbegin() ,
				   channel.rend() ,
				   image_.circularBuffer + image_.bufferPos ) ;
	}
	else {
		std::copy( channel.begin() ,
				   channel.end() ,
				   image_.circularBuffer + image_.bufferPos + image_.width / 2  ) ;

		image_.bufferPos += image_.width ;
	}

	//Control circular buffer
	if ( image_.bufferPos == image_.bufferSize )
		image_.bufferPos = 0 ;

}



void
CImagenexSidescanSonar::imageDraw() {


	//Copy from circularBuffer to imageView
	memcpy( image_.imageView.data ,
			image_.circularBuffer + image_.bufferPos ,
			image_.bufferSize - image_.bufferPos ) ;

	if ( image_.bufferPos > 0 )
		memcpy( image_.imageView.data + ( image_.bufferSize - image_.bufferPos ) ,
				image_.circularBuffer ,
				image_.bufferPos ) ;

	//View Image
	imageAddInfo();
	cv::imshow( "Imaginex Sidescan Sonar", image_.imageView ) ;
	cv::waitKey( 5 ) ;
}



void
CImagenexSidescanSonar::imageScaleChannel( const IIODevice::DataBuffer & channelIn ,
										   IIODevice::DataBuffer & channelOut ) {

	double pxRatio = 2000.0 / static_cast< double >( image_.width ) ;
	double inPos = 0 ;
	int inPosRound = 0 ;
	int outPos = 0 ;

	channelOut.resize( image_.width / 2 ) ;

	while ( outPos < ( image_.width / 2 ) ) {
		inPosRound = round( inPos ) ;
		channelOut[ outPos++ ] = channelIn[ inPosRound ] ;
		inPos += pxRatio ;
	}

}



void
CImagenexSidescanSonar::imageAddInfo() {

	//Add lines
	int dist = ( image_.width / 2 )  / 5 ;
	cv::Point p1,p2 ;

	for ( unsigned char i = 0 ; i < 10 ; i++ ) {
		p1.x = i * dist ;
		p1.y = 0 ;
		p2.x = i * dist ;
		p2.y = image_.height - 1 ;
		cv::line( image_.imageView, p1, p2, CV_RGB(255, 255, 255) ) ;
	}

}



