/*
 * C872File.cpp
 *
 *  Created on: Feb 11, 2011
 *      Author: Joel Vidal
 */

#include "C872File.h"



void
C872File::open( char* filename )
{
	char fName[255];

	strcpy( fName, filename );
	strcat( fName, ".872" );

	outFile_.open( fName );
	pingNumber_ = 0;
}



void
C872File::inputData( char* portData,
					 char* starboardData,
					 Byte freq,
					 Byte range,
					 Byte dataGain,
					 char* headerbuffer )
{

	Byte buffer[5000];
	int nBytes, i, temp;
	char strMonth[4], dateTxt[12], timeTxt[9], uSecondsTxt[5];

	//Structs
	time_t rawtime;
	struct tm * timeinfo;
	struct timeval tv;


	//Posem el buffer a zero;
	for ( i = 0; i < 5000; i++ )
		buffer[i] = 0;


	//872FILE: File Header (1000 Bytes)

	int index = 0;
	//872
	buffer[index++] = '8';
	buffer[index++] = '7';
	buffer[index++] = '2';

	buffer[index++] = 0;//version: 0-current version 1.xx, 1-later version

	pingNumber_++;

	//index=4;
	buffer[index++] = static_cast<Byte>( ( pingNumber_ & 0xFF000000 ) >> 24 );
	buffer[index++] = static_cast<Byte>( ( pingNumber_ & 0x00FF0000 ) >> 16 );
	buffer[index++] = static_cast<Byte>( ( pingNumber_ & 0x0000FF00 ) >> 8 );
	buffer[index++] = static_cast<Byte>( ( pingNumber_ & 0x000000FF ) ); // PingNumber

	nBytes = 4096; //total Bytes for this record, index=8,9
	buffer[index++] = static_cast<Byte>( ( nBytes & 0xFF00 ) >> 8 );
	buffer[index++] = static_cast<Byte>( nBytes & 0x00FF );

	nBytes = 1000; //data points per channel, index=10,11
	buffer[index++] = static_cast<Byte>( ( nBytes & 0xFF00 ) >> 8 );
	buffer[index++] = static_cast<Byte>( nBytes & 0x00FF );

	buffer[index++] = 1;//Bytes per data point 12
	buffer[index++] = 8;//data point bit depth 13
	buffer[index++] = 0;//(Byte)(((gps_type&0x03)<<4)|0x04);//GPS strings type 14

	nBytes = 3000; //GPS string data start address offset 15-16
	buffer[index++] = static_cast<Byte>( ( nBytes & 0xFF00 ) >> 8 );
	buffer[index++] = static_cast<Byte>( nBytes & 0x00FF );

	//Event
	nBytes=0; //Event and annotation 17-18
	buffer[index++] = static_cast<Byte>( ( nBytes & 0xFF00 ) >> 8 );
	buffer[index++] = static_cast<Byte>( nBytes & 0x00FF );

	//get date and time
	time ( &rawtime );
	timeinfo = localtime ( &rawtime );

	int tempmonth = timeinfo->tm_mon;
	switch(tempmonth) {
		case 0 	: sprintf( strMonth, "%s", "JAN" ); break;
		case 1 	: sprintf( strMonth, "%s", "FEB" ); break;
		case 2 	: sprintf( strMonth, "%s", "MAR" ); break;
		case 3 	: sprintf( strMonth, "%s", "APR" ); break;
		case 4 	: sprintf( strMonth, "%s", "MAY" ); break;
		case 5 	: sprintf( strMonth, "%s", "JUN" ); break;
		case 6 	: sprintf( strMonth, "%s", "JUL" ); break;
		case 7 	: sprintf( strMonth, "%s", "AUG" ); break;
		case 8 	: sprintf( strMonth, "%s", "SEP" ); break;
		case 9 	: sprintf( strMonth, "%s", "OCT" ); break;
		case 10 : sprintf( strMonth, "%s", "NOV" ); break;
		case 11 : sprintf( strMonth, "%s", "DEC" ); break;
	}

	//Date DD-MMM-YYYY
	//tm_year anys des del 1900
	sprintf( dateTxt, "%02d-%s-%04d", timeinfo->tm_mday, strMonth, ( timeinfo->tm_year + 1900 ) );
	for( i = 0; i < 11; i++ )
		buffer[index++] = static_cast<Byte>( dateTxt[i] );
	buffer[index++] = '\0';

	//Time HH:MM:SS
	sprintf( timeTxt, "%02d:%02d:%02d", timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec );
	for( i = 0; i < 8; i++ )
		buffer[index++] = static_cast<Byte>( timeTxt[i] );
	buffer[index++] = '\0';

	gettimeofday (&tv, 0);

	sprintf ( uSecondsTxt, ".%03d", ( tv.tv_usec / 1000 ) );
	for(i = 0; i < 4; i++)
		buffer[index++] = static_cast<Byte>( uSecondsTxt[i] );
	buffer[index++] = '\0';

	buffer[index++] = freq;

	Byte rangeIndex;
	switch ( range ) {
	case 10 	: 	rangeIndex = 5; 	break;
	case 20 	: 	rangeIndex = 6;		break;
	case 30 	: 	rangeIndex = 7;		break;
	case 40 	: 	rangeIndex = 8;		break;
	case 50 	: 	rangeIndex = 9;		break;
	case 60 	: 	rangeIndex = 10;	break;
	case 80 	: 	rangeIndex = 11;	break;
	case 100 	: 	rangeIndex = 12;	break;
	case 125 	: 	rangeIndex = 13;	break;
	case 150 	: 	rangeIndex = 14;	break;
	case 200 	: 	rangeIndex = 15;	break;
	default 	: 	rangeIndex = 15;
	}

	buffer[index++] = rangeIndex;
	buffer[index++] = dataGain;
	buffer[index++] = 0;//in order to match with YellowFin

	buffer[index++] = static_cast<Byte>( ( repRate_ & 0xFF00 ) >> 8 );
	buffer[index++] = static_cast<Byte>( repRate_ & 0x00FF );

	nBytes = 1500*10; //sound velocity
	buffer[index++] = static_cast<Byte>( ( nBytes & 0xFF00 ) >> 8 );
	buffer[index++] = static_cast<Byte>( nBytes & 0x00FF );

	for ( i = 0; i < 12; i++ )	//start at index=53
		buffer[index + i] = static_cast<Byte>( headerbuffer[i] );//add sonar header 12 Bytes

	index = 70;
	buffer[index++] = 1;//AUV board
	buffer[index++] = range;

	temp = 0;// (UINT)(m_dAUVSpeed*100);
	buffer[index++] = static_cast<Byte>( ( temp & 0xFF00 ) >> 8 );
	buffer[index++] = static_cast<Byte>( temp & 0x00FF );

	outFile_.write( ( char* )buffer, 1000 );


	//872FILE: PORT CHANNEL ECHO DATA
	char portDataR[1000];
	for (i = 0; i < 1000; i++) portDataR[i] = portData[999-i];
	
	outFile_.write( portDataR, 1000 );


	//872FILE: STARBOARD CHANNEL ECHO DATA
	outFile_.write( starboardData, 1000 );


	//872FILE: zero fill for GPS
	for( i = 0; i < 1001; i++ )
		buffer[i] = 0;//init to all zeros

	outFile_.write( ( char* )buffer, 1000 );


	//872FILE: Zero fill & pointer to a previous ping Bytes of this ping + last ping

	//************** Zero fill ********************
	//	for(i=0;i<1001;i++)buffer[i]=0;//init to all zeros
	//4094-4095
	//pointer to a previous ping Bytes of this ping + last ping
	nBytes = 4096 + 4096;
	buffer[94] = static_cast<Byte>( ( nBytes & 0xFF00 ) >> 8 );
	buffer[95] = static_cast<Byte>( nBytes & 0x00FF );

	outFile_.write( ( char* )buffer, 96 );

	//RepRate
	gettimeofday( &tv, 0 );
	currentTime_ = ( static_cast<double>( tv.tv_sec ) * 1000.0 ) + ( static_cast<double>( tv.tv_usec ) / 1000.0 );
	repRate_ = abs( static_cast<double>( currentTime_ - lastTime_ ) );
	lastTime_ = currentTime_;


	outFile_.flush(); //Update file
}



void C872File::close(){
	outFile_.close();
}



C872File::C872File()
{
}



C872File::~C872File()
{
}

